// COPYRIGHT LICENSE:  This information contains sample code provided in source code form.
// You may copy, modify, and distribute these sample programs in any form without payment
// to IBM for the purposes of developing, using, marketing or distributing application
// programs conforming to the application programming interface for the operating platform
// for which the sample code is written. Notwithstanding anything to the contrary,
// IBM PROVIDES THE SAMPLE SOURCE CODE ON AN "AS IS" BASIS AND IBM DISCLAIMS ALL WARRANTIES,
// EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OR CONDITIONS
// OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE, AND
// ANY WARRANTY OR CONDITION OF NON-INFRINGEMENT.  IBM SHALL NOT BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR
// OPERATION OF THE SAMPLE SOURCE CODE.  IBM HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS TO THE SAMPLE SOURCE CODE.
//
// (C) Copyright IBM Corp. 2011, 2012.
// All Rights Reserved. Licensed Materials - Property of IBM.

package com.ibm.ws.batch.ilc.sample;

import java.util.Properties;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.sql.Connection;
import java.sql.SQLException;
import javax.sql.DataSource;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import com.ibm.websphere.batch.BatchJobStepInterface;
import com.ibm.websphere.batch.BatchConstants;
import com.ibm.websphere.batch.BatchDataStreamMgr;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.context.JobStepContextMgr;

import com.ibm.websphere.batch.ilc.ILContainerFactory;
import com.ibm.websphere.batch.ilc.ILContainer;
import com.ibm.websphere.batch.ilc.ILContainerException;
import com.ibm.websphere.batch.ilc.ILProcedureException;
import com.ibm.ws.batch.ilc.sample.COBEMP;
import com.ibm.ws.batch.ilc.sample.HOLDCURS;
import com.ibm.ws.batch.ilc.sample.parameters.ILCEMPREC;
import com.ibm.ws.batch.ilc.sample.parameters.HC_OP;
import com.ibm.ws.batch.ilc.sample.ILCEmpRecord;
import com.ibm.ws.batch.ilc.sample.bds.ILCHoldableCursorBDS;

public class ILCEmpProcessor implements BatchJobStepInterface
{
  private static final String _cName = ILCEmpProcessor.class.getName();
  private static Logger logger = Logger.getLogger(_cName);

  public static final String SHARED_DB2_DS_JNDI_NAME = "shared_db2_ds_jndi_name";
  public static final String MAX_RECORD_COUNT = "max_record_count";

  private Properties _props = null;
  private int _maxRecordCount = 0;
  private String _sharedDb2JndiName = null;
  private int _recordI = 0;

  private AbstractBatchDataInputStream _inputEmpBDS;
  private AbstractBatchDataOutputStream _outputEmpBDS;

  private COBEMP _cobemp = null;
  private HOLDCURS _holdcurs = null;
  private Connection _sharedDb2Conn = null;

  private ILContainer _ilc = null;

  /**
   * createJobStep performs the following steps:
   * 1. read the job-step properties.
   * 2. create the ILContainer.
   * 3. create the BDSs.
   * 4. create the COBOL call stub(s).
   */
  public void createJobStep()
  {
    final String mName = "createJobStep";
    logger.entering(_cName, mName);

    _maxRecordCount = Integer.parseInt(_props.getProperty(MAX_RECORD_COUNT,"0"));
    _sharedDb2JndiName = _props.getProperty(SHARED_DB2_DS_JNDI_NAME);

    logger.fine(MAX_RECORD_COUNT + ": " + _maxRecordCount);
    logger.fine(SHARED_DB2_DS_JNDI_NAME +": " + _sharedDb2JndiName);

    try
    {
      // Create the Java-COBOL Container
      if (_sharedDb2JndiName != null)
      {
        // Attach the DB2 connection to the container.
        _sharedDb2Conn = getDB2Connection(_sharedDb2JndiName);
        _ilc = ILContainerFactory.getFactory().create(_sharedDb2Conn);

        // Create the HOLDCURS for use with the ILCHoldableCursorBDS custom BDS.
        // The ILCHoldableCursorBDS retrieves EMP records using a WITH HOLD CURSOR
        // in Cobol, by way of the shared DB2 connection.
        _holdcurs = new HOLDCURS();
      }
      else
      {
        _ilc = ILContainerFactory.getFactory().create();
      }

      // Open BDS
      _inputEmpBDS= (AbstractBatchDataInputStream)BatchDataStreamMgr.getBatchDataStream("inputEmp", getJobStepID());
      _outputEmpBDS = (AbstractBatchDataOutputStream)BatchDataStreamMgr.getBatchDataStream("outputEmp", getJobStepID());

      // If we're using the ILCHoldableCursorBDS custom BDS, set the ILContainer info into it.
      if (_inputEmpBDS instanceof ILCHoldableCursorBDS)
        ((ILCHoldableCursorBDS)_inputEmpBDS).setILContainer(_ilc,_holdcurs);
      if (_outputEmpBDS instanceof ILCHoldableCursorBDS)
        ((ILCHoldableCursorBDS)_outputEmpBDS).setILContainer(_ilc,_holdcurs);

      // Create the COBOL call stub
      _cobemp = new COBEMP();
    }
    catch (BatchContainerDataStreamException e)
    {
      logger.log(Level.FINE,"Caught BatchContainerDataStreamException opening BDS",e);
      throw new RuntimeException(e);
    }
    catch (ILContainerException ilce)
    {
      logger.log(Level.FINE,"Caught ILContainerException",ilce);
      throw new RuntimeException(ilce);
    }
    catch (Throwable t)
    {
      logger.log(Level.FINE,"Caught Throwable opening BDS",t);
      throw new RuntimeException(t);
    }
    logger.exiting(_cName, mName);
  }



  /**
   * If we previously opened a shared DB2 connection, close it here.
   */
  public int destroyJobStep()
  {
    final String mName = "destroyJobStep";
    logger.entering(_cName, mName);

    if (_sharedDb2Conn != null)
    {
      try
      {
        logger.fine("closing shared DB2 connection");
        _sharedDb2Conn.close();
      }
      catch (SQLException e)
      {
        logger.log(Level.FINE,"Caught SQLException",e);
        throw new RuntimeException(e);
      }
    }

    logger.exiting(_cName, mName);
    return 0;
  }

  public Properties getProperties()
  {
    return _props;
  }

  /**
   * Process ILC EMP records.
   * Read the record from the input BDS.
   * Pass the record to cobol (COBEMP) for processing.
   * Write the record to the output BDS.
   */
  public int processJobStep()
  {
    final String mName = "processJobStep";
    _recordI++;

    logger.entering(_cName,mName, "record count: " + _recordI);

    try
    {
      if (_inputEmpBDS.hasNext())
      {
        ILCEmpRecord ilcEmpRecord = (ILCEmpRecord)_inputEmpBDS.read();
        logger.fine("read record from inputEmpBDS: " + ilcEmpRecord);

        ilcEmpRecord.writeToILCEMPREC(_cobemp.getILCEMPREC());
        printILCEMPREC(_cobemp.getILCEMPREC(),"ILCEMPREC prior to invoke");

        logger.fine("Invoking COBEMP...");
        _ilc.invokeProcedure(_cobemp);

        printILCEMPREC(_cobemp.getILCEMPREC(),"ILCEMPREC after invoke");
        ilcEmpRecord.readFromILCEMPREC(_cobemp.getILCEMPREC());
        logger.fine("updated ILCEmpRecord: " + ilcEmpRecord);

        _outputEmpBDS.write(ilcEmpRecord);
      }
      else
      {
        logger.fine("inputEmp BDS EXHAUSTED. job done.");
        return BatchConstants.STEP_COMPLETE;
      }
    }
    catch (ILProcedureException ilpe)
    {
      logger.log(Level.FINE,"Caught ILProcedureException",ilpe);
      return BatchConstants.STEP_COMPLETE_EXECUTION_FAILED;
    }
    catch (ILContainerException ilce)
    {
      logger.log(Level.FINE,"Caught ILContainerException",ilce);
      return BatchConstants.STEP_COMPLETE_EXECUTION_FAILED;
    }
    catch (Exception e)
    {
      logger.log(Level.FINE,"caught exception reading/writing BDS", e);
      throw new RuntimeException(e);
    }

    if (_maxRecordCount > 0 && _recordI >= _maxRecordCount)
      return BatchConstants.STEP_COMPLETE;
    else
      return BatchConstants.STEP_CONTINUE;

  }

  public void setProperties(Properties props)
  {
    _props = props;
  }

  private void printILCEMPREC(ILCEMPREC ilcemprec, String msg)
  {
    logger.fine(msg + " " +  ilcemprec.getEmpno() + ":" +
                             ilcemprec.getFirstnme__len() + ":" +
                             ilcemprec.getFirstnme__name() + ":" +
                             ilcemprec.getLastname__len() + ":" +
                             ilcemprec.getLastname__name() + ":" +
                             ilcemprec.getEdlevel() + ":" +
                             ilcemprec.getSalary() + ":" + 
                             ilcemprec);
  }

  private String getJobStepID()
  {
    String retMe = JobStepContextMgr.getContext().getJobStepID().getJobstepid();
    logger.fine("jobstepid: " + retMe);
    return retMe;
  }


  /**
   * Retrieve a DB2 connection from the DataSource identified by the given jndiName.
   */
  private Connection getDB2Connection(String jndiName)
    throws NamingException, SQLException
  {
    final String mName = "getDB2Connection";
    logger.entering(_cName, mName, jndiName);

    InitialContext ic = new InitialContext();
    DataSource datasourceType2 = (DataSource) ic.lookup(jndiName);
    Connection con = datasourceType2.getConnection(); 
    logger.fine("DataSource " + jndiName + " was found and a connection was established.");

    logger.exiting(_cName, mName, con);
    return con;
  }

}
