// COPYRIGHT LICENSE:  This information contains sample code provided in source code form.
// You may copy, modify, and distribute these sample programs in any form without payment
// to IBM for the purposes of developing, using, marketing or distributing application
// programs conforming to the application programming interface for the operating platform
// for which the sample code is written. Notwithstanding anything to the contrary,
// IBM PROVIDES THE SAMPLE SOURCE CODE ON AN "AS IS" BASIS AND IBM DISCLAIMS ALL WARRANTIES,
// EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OR CONDITIONS
// OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE, AND
// ANY WARRANTY OR CONDITION OF NON-INFRINGEMENT.  IBM SHALL NOT BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR
// OPERATION OF THE SAMPLE SOURCE CODE.  IBM HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS TO THE SAMPLE SOURCE CODE.
//
// (C) Copyright IBM Corp. 2011, 2012.
// All Rights Reserved. Licensed Materials - Property of IBM.

package com.ibm.ws.batch.ilc.sample.bds;

import java.util.Properties;
import java.util.logging.Logger;

import com.ibm.websphere.batch.ilc.ILContainerFactory;
import com.ibm.websphere.batch.ilc.ILContainer;
import com.ibm.websphere.batch.ilc.ILContainerException;
import com.ibm.websphere.batch.ilc.ILProcedureException;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

import com.ibm.ws.batch.ilc.sample.HOLDCURS;
import com.ibm.ws.batch.ilc.sample.parameters.ILCEMPREC;
import com.ibm.ws.batch.ilc.sample.parameters.HC_OP;
import com.ibm.ws.batch.ilc.sample.parameters.HC_RC;
import com.ibm.ws.batch.ilc.sample.ILCEmpRecord;
import com.ibm.ws.batch.ilc.sample.ILCEmpProcessor;

public class ILCHoldableCursorBDS extends AbstractBatchDataStream implements
        AbstractBatchDataInputStream, AbstractBatchDataOutputStream
{
  private static final String _cName = ILCHoldableCursorBDS.class.getName();
  private static Logger logger = Logger.getLogger(_cName);

  private static final int HC_OP_OPENCURSOR = 1;
  private static final int HC_OP_FETCHCURSOR = 2;
  private static final int HC_OP_CLOSECURSOR = 3;
  private static final int HC_OP_UPDATECURSOR = 4;

  private static final int SQLCODE_CURSOR_NOT_OPEN = -501;
  private static final int SQLCODE_CURSOR_EXHAUSTED = 100; 

  private ILContainer _ilc = null;
  private HOLDCURS _holdcurs = null;
  private boolean _prefetched = false;
  private ILCEmpRecord _prefetchedRecord = null;
  private Properties _props = null;

  public void setILContainer(ILContainer ilc, HOLDCURS holdcurs)
  {
    _ilc = ilc;
    _holdcurs = holdcurs;

    if (_holdcurs == null)
      throw new IllegalArgumentException("HOLDCURS is null.  Make sure " + ILCEmpProcessor.SHARED_DB2_DS_JNDI_NAME + " is set in the step properties");
  }

  public void close() throws BatchContainerDataStreamException
  {
    // Nothing to do here.
    // The cursor is closed when the DB2 connection is closed.
  }

  public String externalizeCheckpointInformation()
  {
    // No checkpointing.
    return null;
  }

  protected void initialize(Properties props)
  {
    _props = props;
  }

  public void internalizeCheckpointInformation(String token) throws RuntimeException
  {
    // No checkpointing.
  }

  public void open() throws BatchContainerDataStreamException
  {
    // Don't want to open the cursor here, since this method is called under 
    // a different transaction context (local) than processJobStep (global).
  }

  public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException
  {
    // No checkpointing.
  }

  public Object fetchHeader()
  {
    // Nothing to do here.
    return null;
  }

  public boolean hasNext() throws Exception
  {
    if (!_prefetched)
    {
      _prefetchedRecord = fetchCursor();
      _prefetched = true;
    }

    return (_prefetchedRecord != null);
  }

  public Object read() throws Exception
  {
    ILCEmpRecord ilcEmpRecord = null;
    if (_prefetched)
    {
      ilcEmpRecord = _prefetchedRecord;
      _prefetchedRecord = null;
      _prefetched = false;
    }
    else
    {
      ilcEmpRecord = fetchCursor();
    }

    return ilcEmpRecord;
  }

  public Object read(Object parameters) throws Exception
  {
    // Nothing to do here.
    return null;
  }

  public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException
  {
    // No checkpointing.
  }


  /** 
   * Updates the record at the current cursor position.
   */
	public void write(Object record) throws Exception
  {
    updateCursor((ILCEmpRecord)record);
  }
	
	public  void flush() throws Exception
  {
    // Nothing to do here.
  }
	
	public void writeHeader(Object record) throws Exception
  {
    // Nothing to do here.
  }

  private int openCursor()
    throws ILContainerException, ILProcedureException
  {
    _holdcurs.getHC_OP().setHc__op(HC_OP_OPENCURSOR);
    int rc = _ilc.invokeProcedure(_holdcurs);
    int hc_rc = _holdcurs.getReturnValue().getHc__rc();
    logger.fine("openCursor: " + hc_rc + ":" + rc );
    return hc_rc;
  }

  private ILCEmpRecord fetchCursor()
    throws ILContainerException, ILProcedureException
  {
    logger.entering(_cName,"fetchCursor");
    ILCEmpRecord ilcEmpRecord = null;

    // Invoke HOLDCURS to fetch the next row.
    _holdcurs.getHC_OP().setHc__op(HC_OP_FETCHCURSOR);
    int rc = _ilc.invokeProcedure(_holdcurs);
    int hc_rc = _holdcurs.getReturnValue().getHc__rc();

    if (hc_rc == 0) // success.
    {
      ilcEmpRecord = new ILCEmpRecord(_holdcurs.getILCEMPREC());
    }
    else if (hc_rc == SQLCODE_CURSOR_NOT_OPEN)
    {
      logger.fine("received CURSOR_NOT_OPEN error: " + hc_rc + ":" + rc + ". Opening cursor");
      rc = openCursor();
      if (rc == 0)  // success, recurse.
        return fetchCursor();
    }
    else if (hc_rc == SQLCODE_CURSOR_EXHAUSTED)
    {
      logger.fine("CURSOR exhausted: " + hc_rc + ":" + rc );
    }
    else
    {
      logger.fine("unrecognized SQLCODE: " + hc_rc + ":" + rc );
    }

    logger.exiting(_cName,"fetchCursor", ilcEmpRecord);
    return ilcEmpRecord;
  }


  private int updateCursor(ILCEmpRecord ilcEmpRecord)
    throws ILContainerException, ILProcedureException
  {
    _holdcurs.getHC_OP().setHc__op(HC_OP_UPDATECURSOR);
    ilcEmpRecord.writeToILCEMPREC(_holdcurs.getILCEMPREC());

    int rc = _ilc.invokeProcedure(_holdcurs);
    int hc_rc = _holdcurs.getReturnValue().getHc__rc();

    logger.fine("updateCursor: " + hc_rc + ":" + rc );
    return hc_rc;
  }


  private int closeCursor()
    throws ILContainerException, ILProcedureException
  {
    _holdcurs.getHC_OP().setHc__op(HC_OP_CLOSECURSOR);
    int rc = _ilc.invokeProcedure(_holdcurs);
    int hc_rc = _holdcurs.getReturnValue().getHc__rc();
    logger.fine("closeCursor: " + hc_rc + ":" + rc );
    return hc_rc;
  }

}
